DEVICE_REGISTER = "api/mobile/register/new"
STUDENT_LIST = "api/mobile/register/hebe"

DATA_ROOT = "api/mobile"
DATA_BY_PUPIL = "byPupil"
DATA_BY_PERSON = "byPerson"
DATA_BY_PERIOD = "byPeriod"
DATA_DELETED = "deleted"

DATA_ADDRESSBOOK = "addressbook"
DATA_INTERNAL_TIME = "internal/time"
DATA_LUCKY_NUMBER = "school/lucky"

DATA_EXAM = "exam"
DATA_ATTENDANCE = "lesson"
DATA_GRADE = "grade"
DATA_GRADE_SUMMARY = "grade/summary"
DATA_GRADE_AVERAGE = "grade/average"
DATA_HOMEWORK = "homework"
DATA_MESSAGE = "message"
DATA_TIMETABLE = "schedule"
DATA_TIMETABLE_CHANGES = "schedule/changes"



# -*- coding: utf-8 -*-

from related import StringField, immutable

from ._serializable import Serializable


@immutable
class Attachment(Serializable):
    """An attachment

    :var str ~.name: Name
    :var str ~.link: Link
    """

    name: str = StringField(key="Name")
    link: str = StringField(key="Link")



# -*- coding: utf-8 -*-
from typing import AsyncIterator, List, Union

from related import (
    BooleanField,
    ChildField,
    IntegerField,
    SequenceField,
    StringField,
    immutable,
)

from .._api_helper import FilterType
from .._endpoints import DATA_HOMEWORK
from ..model import Attachment, DateTime, Serializable, Subject, Teacher


@immutable
class Homework(Serializable):
    """A homework.

    :var int ~.id: homework's external ID
    :var str ~.key: homework's key (UUID)
    :var int ~.homework_id: homework's internal ID
    :var str ~.content: homework's content
    :var `~vulcan.hebe.model.DateTime` ~.date_created: homework's creation date
    :var `~vulcan.hebe.model.Teacher` ~.creator: the teacher who added
        the homework
    :var `~vulcan.hebe.model.Subject` ~.subject: the homework's subject
    :var List[Attachment] ~.attachments: attachments added to homework
    :var bool ~.is_answer_required: Is an answer required
    :var `~vulcan.hebe.model.DateTime` ~.deadline: homework's date and time
    :var `~vulcan.hebe.model.DateTime` ~.answer_deadline: homework's answer deadline
    :var `~vulcan.hebe.model.DateTime` ~.answer_date: homework's answer date and time
    """

    id: int = IntegerField(key="Id")
    key: str = StringField(key="Key")
    homework_id: int = StringField(key="IdHomework")
    content: str = StringField(key="Content")
    date_created: DateTime = ChildField(DateTime, key="DateCreated")
    creator: Teacher = ChildField(Teacher, key="Creator")
    subject: Subject = ChildField(Subject, key="Subject")
    attachments: List[Attachment] = SequenceField(
        Attachment, key="Attachments", repr=True
    )
    is_answer_required: Subject = BooleanField(key="IsAnswerRequired")
    deadline: DateTime = ChildField(DateTime, key="Deadline")
    answer_deadline: DateTime = ChildField(
        DateTime, key="AnswerDeadline", required=False
    )
    answer_date: DateTime = ChildField(DateTime, key="AnswerDate", required=False)

    @classmethod
    async def get(
        cls, api, last_sync, deleted, **kwargs
    ) -> Union[AsyncIterator["Homework"], List[int]]:
        """
        :rtype: Union[AsyncIterator[:class:`~vulcan.hebe.data.Homework`], List[int]]
        """
        data = await api.helper.get_list(
            DATA_HOMEWORK,
            FilterType.BY_PUPIL,
            deleted=deleted,
            last_sync=last_sync,
            **kwargs
        )

        for homework in data:
            yield Homework.load(homework)



# -*- coding: utf-8 -*-

import logging
import math
import platform
import time
import urllib
import uuid as _uuid
from datetime import datetime

import aiohttp

from ._exceptions import InvalidTokenException

APP_NAME = "DzienniczekPlus 2.0"
APP_VERSION = "1.4.2"
APP_OS = "Android"
APP_USER_AGENT = "Dart/2.10 (dart:io)"

log = logging.getLogger("client")
log.setLevel(logging.INFO)

handler = logging.StreamHandler()
log.addHandler(handler)

TIME_FORMAT_H_M = "%H:%M"


def default_device_model():
    return "Vulcan API (Python {})".format(platform.python_version())


async def get_base_url(token):
    code = token[0:3]
    components = await get_components()
    try:
        return components[code]
    except KeyError:
        raise InvalidTokenException()


async def get_components():
    log.info("Getting Vulcan components...")
    async with aiohttp.ClientSession() as session:
        async with session.get(
            "http://komponenty.vulcan.net.pl/UonetPlusMobile/RoutingRules.txt"
        ) as r:
            if r.headers["Content-Type"] == "text/plain":
                r_txt = await r.text()
                components = (c.split(",") for c in r_txt.split())
                components = {a[0]: a[1] for a in components}
            else:
                components = {}
            components.update({"FK1": "http://api.fakelog.tk"})
            return components


async def get_firebase_token():
    async with aiohttp.ClientSession() as session:
        log.info("Getting Firebase token...")
        aid = "4609707972546570896:3626695765779152704"
        device = aid.split(":")[0]
        app = "pl.edu.vulcan.hebe"

        data = {
            "sender": "987828170337",
            "X-scope": "*",
            "X-gmp_app_id": "1:987828170337:android:ac97431a0a4578c3",
            "app": app,
            "device": device,
        }

        headers = {
            "Authorization": "AidLogin {}".format(aid),
            "User-Agent": "Android-GCM/1.5",
            "app": app,
        }

        async with session.post(
            "https://android.clients.google.com/c2dm/register3",
            data=data,
            headers=headers,
        ) as r:
            r_txt = await r.text()
            return r_txt.split("=")[1]


def millis():
    return math.floor(time.time() * 1000)


def now_datetime():  # UTC+0
    return datetime.utcnow()


def now_iso(dt=None):  # ISO 8601, local timezone
    return (dt or datetime.now()).strftime("%Y-%m-%d %H:%M:%S")


def now_gmt(dt=None):  # RFC 2822, UTC+0
    return (dt or datetime.utcnow()).strftime("%a, %d %b %Y %H:%M:%S GMT")


def uuid(seed=None):
    if seed:
        return str(_uuid.uuid5(_uuid.NAMESPACE_X500, str(seed)))
    return str(_uuid.uuid4())


def urlencode(s):
    return urllib.parse.quote(str(s))



# -*- coding: utf-8 -*-
from datetime import date, datetime
from typing import AsyncIterator, List, Union

from ._api import Api
from .data import (
    Addressbook,
    Attendance,
    ChangedLesson,
    Exam,
    Grade,
    Homework,
    Lesson,
    LuckyNumber,
    Message,
)
from .model import DateTime


class VulcanData:
    """A data client for the API.

    Contains methods for getting all data objects, some in
    form of a list, others as an object. All the methods
    are asynchronous. Additionally, the list getting methods
    return an `AsyncIterator` of the items.

    The data client shall not be constructed outside of the main
    API class.
    """

    def __init__(self, api: Api):
        self._api = api

    async def get_time(self) -> DateTime:
        """Gets the current server time.

        :rtype: :class:`~vulcan.model.DateTime`
        """
        return await DateTime.get(self._api)

    async def get_lucky_number(self, day: date = None) -> LuckyNumber:
        """Gets the lucky number for the specified date.

        :param `datetime.date` day: date of the lucky number to get.
            Defaults to ``None`` (today).
        :rtype: :class:`~vulcan.data.LuckyNumber`
        """
        return await LuckyNumber.get(self._api, day or date.today())

    async def get_addressbook(
        self, **kwargs
    ) -> Union[AsyncIterator[Addressbook], List[int]]:
        """Yields the addressbook.

        :rtype: Union[AsyncIterator[:class:`~vulcan.data.Addressbook`], List[int]]
        """
        return Addressbook.get(self._api, **kwargs)

    async def get_messages(
        self, last_sync: datetime = None, folder=1, **kwargs
    ) -> Union[AsyncIterator[Message], List[int]]:
        """Yields messages received by the currently registered account.

        :param `datetime.datetime` last_sync: date of the last sync,
            gets only the objects updated since this date
        :param int folder: message folder: 1 - received; 2 - sent; 3 - deleted
        :rtype: Union[AsyncIterator[:class:`~vulcan.data.Message`], List[int]]
        """
        return Message.get(self._api, last_sync, folder, **kwargs)

    async def get_grades(
        self, last_sync: datetime = None, deleted=False, **kwargs
    ) -> Union[AsyncIterator[Grade], List[int]]:
        """Yields the student's grades.

        :param `datetime.datetime` last_sync: date of the last sync,
            gets only the objects updated since this date
        :param bool deleted: whether to only get the deleted item IDs
        :rtype: Union[AsyncIterator[:class:`~vulcan.data.Grade`], List[int]]
        """
        return Grade.get(self._api, last_sync, deleted, **kwargs)

    async def get_exams(
        self, last_sync: datetime = None, deleted=False, **kwargs
    ) -> Union[AsyncIterator[Grade], List[int]]:
        """Yields the student's exams.

        :param `datetime.datetime` last_sync: date of the last sync,
            gets only the objects updated since this date
        :param bool deleted: whether to only get the deleted item IDs
        :rtype: Union[AsyncIterator[:class:`~vulcan.data.Exam`], List[int]]
        """
        return Exam.get(self._api, last_sync, deleted, **kwargs)

    async def get_homework(
        self, last_sync: datetime = None, deleted=False, **kwargs
    ) -> Union[AsyncIterator[Homework], List[int]]:
        """Yields the student's homework.

        :param `datetime.datetime` last_sync: date of the last sync,
            gets only the objects updated since this date
        :param bool deleted: whether to only get the deleted item IDs
        :rtype: Union[AsyncIterator[:class:`~vulcan.data.Homework`], List[int]]
        """
        return Homework.get(self._api, last_sync, deleted, **kwargs)

    async def get_lessons(
        self,
        last_sync: datetime = None,
        deleted=False,
        date_from=None,
        date_to=None,
        **kwargs,
    ) -> Union[AsyncIterator[Lesson], List[int]]:
        """Yields the student's lessons.

        :param `datetime.datetime` last_sync: date of the last sync,
            gets only the objects updated since this date
        :param bool deleted: whether to only get the deleted item IDs
        :param `datetime.date` date_from: Date, from which to fetch lessons, if not provided
            it's using the today date (Default value = None)
        :param `datetime.date` date_to: Date, to which to fetch lessons, if not provided
            it's using the `date_from` date (Default value = None)
        :rtype: Union[AsyncIterator[:class:`~vulcan.data.Lesson`], List[int]]
        """
        return Lesson.get(self._api, last_sync, deleted, date_from, date_to, **kwargs)

    async def get_changed_lessons(
        self,
        last_sync: datetime = None,
        deleted=False,
        date_from=None,
        date_to=None,
        **kwargs,
    ) -> Union[AsyncIterator[ChangedLesson], List[int]]:
        """Yields the student's changed lessons.

        :param `datetime.datetime` last_sync: date of the last sync,
            gets only the objects updated since this date
        :param bool deleted: whether to only get the deleted item IDs
        :param `datetime.date` date_from: Date, from which to fetch lessons, if not provided
            it's using the today date (Default value = None)
        :param `datetime.date` date_to: Date, to which to fetch lessons, if not provided
            it's using the `date_from` date (Default value = None)
        :rtype: Union[AsyncIterator[:class:`~vulcan.data.ChangedLesson`], List[int]]
        """
        return ChangedLesson.get(
            self._api, last_sync, deleted, date_from, date_to, **kwargs
        )

    async def get_attendance(
        self,
        last_sync: datetime = None,
        deleted=False,
        date_from=None,
        date_to=None,
        **kwargs,
    ) -> Union[AsyncIterator[Attendance], List[int]]:
        """Fetches attendance from the given date

        :param `datetime.datetime` last_sync: date of the last sync,
            gets only the objects updated since this date
        :param bool deleted: whether to only get the deleted item IDs
        :param `datetime.date` date_from: Date, from which to fetch attendance, if not provided
            it's using the today date (Default value = None)
        :param `datetime.date` date_to: Date, to which to fetch attendance, if not provided
            it's using the `date_from` date (Default value = None)
        :rtype: Union[AsyncIterator[:class:`~vulcan.data.Attendance`], List[int]]
        """
        return Attendance.get(
            self._api, last_sync, deleted, date_from, date_to, **kwargs
        )



class VulcanAPIException(Exception):
    pass


class InvalidTokenException(VulcanAPIException):
    pass


class InvalidPINException(VulcanAPIException):
    pass


class InvalidSymbolException(VulcanAPIException):
    pass


class ExpiredTokenException(VulcanAPIException):
    pass


class UnauthorizedCertificateException(VulcanAPIException):
    pass


class InvalidSignatureValuesException(VulcanAPIException):
    pass



# -*- coding: utf-8 -*-

import json
from typing import Union

import aiohttp
from uonet_request_signer_hebe import get_signature_values
from yarl import URL

from ._api_helper import ApiHelper
from ._exceptions import (
    ExpiredTokenException,
    InvalidPINException,
    InvalidSignatureValuesException,
    InvalidSymbolException,
    InvalidTokenException,
    UnauthorizedCertificateException,
    VulcanAPIException,
)
from ._keystore import Keystore
from ._utils import (
    APP_NAME,
    APP_OS,
    APP_USER_AGENT,
    APP_VERSION,
    log,
    millis,
    now_datetime,
    now_gmt,
    now_iso,
    urlencode,
    uuid,
)
from .model import Period, Student


class Api:
    """The API service class.

    Provides methods for sending GET/POST requests on a higher
    level, automatically generating the required headers
    and other values.

    :var `~vulcan._api_helper.ApiHelper` ~.helper: a wrapper for getting
         most data objects more easily
    """

    def __init__(self, keystore: Keystore, account=None, session=None):
        self._session = aiohttp.ClientSession() if not session else session
        # if not isinstance(keystore, Keystore):
        #     raise ValueError("The argument must be a Keystore")
        self._keystore = keystore
        if account:
            self._account = account
            self._rest_url = account.rest_url
        self._student = None
        self._period = None
        self.helper = ApiHelper(self)

    def _build_payload(self, envelope: dict) -> dict:
        return {
            "AppName": APP_NAME,
            "AppVersion": APP_VERSION,
            "CertificateId": self._keystore.fingerprint,
            "Envelope": envelope,
            "FirebaseToken": self._keystore.firebase_token,
            "API": 1,
            "RequestId": uuid(),
            "Timestamp": millis(),
            "TimestampFormatted": now_iso(),
        }

    def _build_headers(self, full_url: str, payload: str) -> dict:
        dt = now_datetime()
        digest, canonical_url, signature = get_signature_values(
            self._keystore.fingerprint,
            self._keystore.private_key,
            payload,
            full_url,
            dt,
        )

        headers = {
            "User-Agent": APP_USER_AGENT,
            "vOS": APP_OS,
            "vDeviceModel": self._keystore.device_model,
            "vAPI": "1",
            "vDate": now_gmt(dt),
            "vCanonicalUrl": canonical_url,
            "Signature": signature,
        }

        if digest:
            headers["Digest"] = digest
            headers["Content-Type"] = "application/json"

        return headers

    async def _request(
        self, method: str, url: str, body: dict = None, **kwargs
    ) -> Union[dict, list]:
        if self._session.closed:
            raise RuntimeError("The AioHttp session is already closed.")

        full_url = (
            url
            if url.startswith("http")
            else self._rest_url + url
            if self._rest_url
            else None
        )
        if not full_url:
            raise ValueError("Relative URL specified but no account loaded")

        payload = self._build_payload(body) if body and method == "POST" else None
        payload = json.dumps(payload) if payload else None
        headers = self._build_headers(full_url, payload)

        log.debug(" > {} to {}".format(method, full_url))

        # a workaround for aiohttp incorrectly re-encoding the full URL
        full_url = URL(full_url, encoded=True)

        async with self._session.request(
            method, full_url, data=payload, headers=headers, **kwargs
        ) as r:
            try:
                response = await r.json()
                status = response["Status"]
                envelope = response["Envelope"]

                # check for the presence of a b64 string preceded with ': '
                if status["Code"] == 100 and ": " in status["Message"]:
                    raise InvalidSignatureValuesException()

                elif status["Code"] == 108:
                    log.debug(" ! " + str(status))
                    raise UnauthorizedCertificateException()

                elif status["Code"] == 200:
                    log.debug(" ! " + str(status))
                    raise InvalidTokenException()

                elif status["Code"] == 203:
                    log.debug(" ! " + str(status))
                    raise InvalidPINException()

                elif status["Code"] == 204:
                    log.debug(" ! " + str(status))
                    raise ExpiredTokenException()

                elif status["Code"] == -1:
                    log.debug(" ! " + str(status))
                    raise InvalidSymbolException()

                elif status["Code"] != 0:
                    log.debug(" ! " + str(status))
                    raise VulcanAPIException(status["Message"])

                log.debug(" < " + str(envelope))
                return envelope
            except ValueError:
                raise VulcanAPIException("An unexpected exception occurred.")

    async def get(self, url: str, query: dict = None, **kwargs) -> Union[dict, list]:
        query = (
            "&".join(x + "=" + urlencode(query[x]) for x in query) if query else None
        )
        if query:
            url += "?" + query
        return await self._request("GET", url, body=None, **kwargs)

    async def post(self, url: str, body: dict, **kwargs) -> Union[dict, list]:
        return await self._request("POST", url, body, **kwargs)

    async def open(self):
        if self._session.closed:
            self._session = aiohttp.ClientSession()

    async def close(self):
        await self._session.close()

    @property
    def account(self):
        return self._account

    @property
    def student(self) -> Student:
        return self._student

    @student.setter
    def student(self, student: Student):
        if not self._account:
            raise AttributeError("Load an Account first!")
        self._rest_url = self._account.rest_url + student.unit.code + "/"
        self._student = student
        self.period = student.current_period

    @property
    def period(self) -> Period:
        return self._period

    @period.setter
    def period(self, period: Period):
        self._period = period



# -*- coding: utf-8 -*-

from related import IntegerField, StringField, immutable

from ._api import Api
from ._endpoints import DEVICE_REGISTER
from ._keystore import Keystore
from ._utils import APP_OS, get_base_url, log, uuid
from .model import Serializable


@immutable
class Account(Serializable):
    """An account in the e-register.

    :var int ~.login_id: the account's login ID
    :var str ~.user_login: the account's login name (email/username)
    :var str ~.user_name: probably the same as user_login
    :var str ~.rest_url: the API base URL for the partition symbol
    """

    login_id: int = IntegerField(key="LoginId")
    user_login: str = StringField(key="UserLogin")
    user_name: str = StringField(key="UserName")
    rest_url: str = StringField(key="RestURL")

    @staticmethod
    async def register(
        keystore: Keystore, token: str, symbol: str, pin: str
    ) -> "Account":
        token = str(token).upper()
        symbol = str(symbol).lower()
        pin = str(pin)

        body = {
            "OS": APP_OS,
            "DeviceModel": keystore.device_model,
            "Certificate": keystore.certificate,
            "CertificateType": "X509",
            "CertificateThumbprint": keystore.fingerprint,
            "PIN": pin,
            "SecurityToken": token,
            "SelfIdentifier": uuid(keystore.fingerprint),
        }

        base_url = await get_base_url(token)
        full_url = "/".join([base_url, symbol, DEVICE_REGISTER])

        log.info("Registering to {}...".format(symbol))

        api = Api(keystore)
        response = await api.post(full_url, body)
        await api.close()

        log.info("Successfully registered as {}".format(response["UserName"]))

        return Account.load(response)



# -*- coding: utf-8 -*-

from ._attachment import Attachment
from ._datetime import DateTime
from ._period import Period
from ._pupil import Gender, Pupil
from ._school import School
from ._serializable import Serializable
from ._student import Student
from ._subject import Subject
from ._teacher import Teacher
from ._team import TeamClass, TeamVirtual
from ._timeslot import TimeSlot
from ._unit import Unit



# -*- coding: utf-8 -*-
from typing import AsyncIterator, List, Union

from related import ChildField, IntegerField, StringField, immutable

from .._api_helper import FilterType
from .._endpoints import DATA_EXAM
from ..model import DateTime, Serializable, Subject, Teacher, TeamClass, TeamVirtual


@immutable
class Exam(Serializable):
    """An exam or short quiz.

    :var int ~.id: exam's ID
    :var str ~.key: exam's key (UUID)
    :var str ~.type: exam's type
    :var str ~.topic: exam's topic
    :var `~vulcan.model.DateTime` ~.date_created: exam's creation date
    :var `~vulcan.model.DateTime` ~.date_modified: exam's modification date
        (may be the same as ``date_created`` if it was never modified)
    :var `~vulcan.model.DateTime` ~.deadline: exam's date and time
    :var `~vulcan.model.Teacher` ~.creator: the teacher who added
        the exam
    :var `~vulcan.model.Subject` ~.subject: the exam's subject
    :var `~vulcan.model.TeamClass` ~.team_class: the class taking the exam
    :var `~vulcan.model.TeamVirtual` ~.team_virtual: the class distribution
        taking the exam, optional
    """

    id: int = IntegerField(key="Id")
    key: str = StringField(key="Key")
    type: str = StringField(key="Type")
    topic: str = StringField(key="Content")
    date_created: DateTime = ChildField(DateTime, key="DateCreated")
    date_modified: DateTime = ChildField(DateTime, key="DateModify")
    deadline: DateTime = ChildField(DateTime, key="Deadline")
    creator: Teacher = ChildField(Teacher, key="Creator")
    subject: Subject = ChildField(Subject, key="Subject")
    team_class: TeamClass = ChildField(TeamClass, key="Class", required=False)
    team_virtual: TeamVirtual = ChildField(
        TeamVirtual, key="Distribution", required=False
    )

    @classmethod
    async def get(
        cls, api, last_sync, deleted, **kwargs
    ) -> Union[AsyncIterator["Exam"], List[int]]:
        """
        :rtype: Union[AsyncIterator[:class:`~vulcan.data.Exam`], List[int]]
        """
        data = await api.helper.get_list(
            DATA_EXAM,
            FilterType.BY_PUPIL,
            deleted=deleted,
            last_sync=last_sync,
            **kwargs
        )

        for exam in data:
            yield Exam.load(exam)



# -*- coding: utf-8 -*-

import io
import json
from typing import Type, TypeVar

from related import immutable, to_json, to_model

T = TypeVar("T")


@immutable
class Serializable:
    """A base class allowing to (de)serialize objects easily into
    appropriate class variables.
    """

    @property
    def as_json(self) -> str:
        """Serialize the object as a JSON string.

        :rtype: str
        """
        return to_json(self)

    @property
    def as_dict(self) -> dict:
        """Serialize the object as a dictionary.

        :rtype: dict
        """
        return json.loads(self.as_json)

    @classmethod
    def load(cls: Type[T], data) -> T:
        """Deserialize provided ``data`` into an instance of ``cls``.

        The ``data`` parameter may be:

        - a JSON string
        - a dictionary
        - a handle to a file containing a JSON string

        :param data: the data to deserialize
        """
        if not data:
            return None
        if isinstance(data, dict):
            return to_model(cls, data)
        elif isinstance(data, io.IOBase):
            return to_model(cls, json.load(data))
        elif isinstance(data, str):
            return to_model(cls, json.loads(data))
        else:
            raise ValueError("Unknown data type")



# -*- coding: utf-8 -*-
from typing import List

from related import ChildField, SequenceField, StringField, immutable

from .._endpoints import STUDENT_LIST
from ._period import Period
from ._pupil import Pupil
from ._school import School
from ._serializable import Serializable
from ._unit import Unit


@immutable
class Student(Serializable):
    """A student object, along with his school, class and period information

    :var str ~.class_: student class
    :var str ~.symbol: the "partition" symbol - can be a town or county name
    :var str ~.symbol_code: the school unit code - often a 6 digit number
    :var `~vulcan.model.Pupil` ~.pupil: contains the student's IDs,
         names and email
    :var `~vulcan.model.Unit` ~.unit: info about the school unit
         (e.g. several school buildings)
    :var `~vulcan.model.School` ~.school: info about the school
         (a single building of the unit)
    :var List[`~vulcan.model.Period`] ~.periods: a list of
         the student's school year periods
    """

    class_: str = StringField(key="ClassDisplay")
    symbol: str = StringField(key="TopLevelPartition")
    symbol_code: str = StringField(key="Partition")

    pupil: Pupil = ChildField(Pupil, key="Pupil")
    unit: Unit = ChildField(Unit, key="Unit")
    school: School = ChildField(School, key="ConstituentUnit")
    periods: List[Period] = SequenceField(Period, key="Periods")

    @property
    def full_name(self) -> str:
        """Gets the student's full name in "FirstName SecondName LastName" format or  "FirstName LastName" format if
        there is no second name.

        :rtype: str
        """
        return " ".join(
            part
            for part in [
                self.pupil.first_name.strip(),
                self.pupil.second_name.strip() if self.pupil.second_name else None,
                self.pupil.last_name.strip(),
            ]
            if part
        )

    @property
    def current_period(self) -> Period:
        """Gets the currently ongoing period of the student.

        :rtype: :class:`~vulcan.model.Period`
        """
        return next((period for period in self.periods if period.current), None)

    def period_by_id(self, period_id: int) -> Period:
        """Gets a period matching the given period ID.

        :param int period_id: the period ID to look for
        :rtype: :class:`~vulcan.model.Period`
        """
        return next((period for period in self.periods if period.id == period_id), None)

    @classmethod
    async def get(cls, api, **kwargs) -> List["Student"]:
        """
        :rtype: List[:class:`~vulcan.model.Student`]
        """
        data = await api.get(STUDENT_LIST, **kwargs)
        return list(Student.load(student) for student in data)



# -*- coding: utf-8 -*-

from related import IntegerField, StringField, immutable

from ._serializable import Serializable


@immutable
class School(Serializable):
    """A single school building.

    :var int ~.id: school ID
    :var str ~.name: school full name
    :var str ~.short_name: school short name
    :var str ~.address: school address (location)
    """

    id: int = IntegerField(key="Id")
    name: str = StringField(key="Name")
    short_name: str = StringField(key="Short")
    address: str = StringField(key="Address", required=False)



# -*- coding: utf-8 -*-

from related import BooleanField, ChildField, IntegerField, immutable

from ._datetime import DateTime
from ._serializable import Serializable


@immutable
class Period(Serializable):
    """A school year period.

    :var int ~.id: the period ID
    :var int ~.level: a grade/level number
    :var int ~.number: number of the period in the school year
    :var bool ~.current: whether the period is currently ongoing
    :var bool ~.last: whether the period is last in the school year
    :var `~vulcan.model.DateTime` ~.start: the period start datetime
    :var `~vulcan.model.DateTime` ~.end: the period end datetime
    """

    id: int = IntegerField(key="Id")
    level: int = IntegerField(key="Level")
    number: int = IntegerField(key="Number")
    current: bool = BooleanField(key="Current")
    last: bool = BooleanField(key="Last")
    start: DateTime = ChildField(DateTime, key="Start")
    end: DateTime = ChildField(DateTime, key="End")



# -*- coding: utf-8 -*-

from related import IntegerField, StringField, immutable

from ._serializable import Serializable


@immutable
class Subject(Serializable):
    """A school subject.

    :var int ~.id: subject ID
    :var str ~.key: subject's key (UUID)
    :var str ~.name: subject's name
    :var str ~.code: subject's code (e.g. short name or abbreviation)
    :var int ~.position: unknown, yet
    """

    id: int = IntegerField(key="Id")
    key: str = StringField(key="Key")
    name: str = StringField(key="Name")
    code: str = StringField(key="Kod")
    position: int = IntegerField(key="Position")



# -*- coding: utf-8 -*-
from typing import List

from ._api import Api
from ._data import VulcanData
from ._utils import log
from .model import Student


class Vulcan:
    """Vulcan API client.

    Contains methods for getting/setting the current student and for
    setting the logging level. All data is fetched from an instance
    of the :class:`~vulcan._data.VulcanData`, accessible
    using the ``data`` variable.

    :var `~vulcan._data.VulcanData` ~.data: the data client
    """

    def __init__(self, keystore, account, session=None, logging_level: int = None):
        self._api = Api(keystore, account, session)
        self._students = []
        self.data = VulcanData(self._api)

        if logging_level:
            Vulcan.set_logging_level(logging_level)

    async def __aenter__(self):
        await self._api.open()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self._api.close()

    async def close(self):
        await self._api.close()

    async def select_student(self):
        """Load a list of students associated with the account.
        Set the first available student as default for the API.
        """
        students = await self.get_students()
        self.student = students[0] if len(students) > 0 else None

    @staticmethod
    def set_logging_level(logging_level: int):
        """Set the API logging level.

        :param int logging_level: logging level from `logging` module
        """
        log.setLevel(logging_level)

    async def get_students(self, cached=True) -> List[Student]:
        """Gets students assigned to this account.

        :param bool cached: whether to allow returning the cached list
        :rtype: List[:class:`~vulcan.model.Student`]
        """
        if self._students and cached:
            return self._students
        self._students = await Student.get(self._api)
        return self._students

    @property
    def student(self) -> Student:
        """Gets/sets the currently selected student.

        :rtype: :class:`~vulcan.model.Student`
        """
        return self._api.student

    @student.setter
    def student(self, value: Student):
        """Changes the currently selected student.

        :param value: the student to select
        :type value: :class:`~vulcan.model.Student`
        """
        self._api.student = value



# -*- coding: utf-8 -*-

from ._addressbook import Addressbook, Role
from ._attendance import Attendance, PresenceType
from ._exam import Exam
from ._grade import Grade, GradeCategory, GradeColumn
from ._homework import Homework
from ._lesson import ChangedLesson, Lesson, LessonChanges, LessonRoom
from ._lucky_number import LuckyNumber
from ._message import Address, Message



# -*- coding: utf-8 -*-

from related import IntegerField, StringField, immutable

from ._serializable import Serializable


@immutable
class Unit(Serializable):
    """A group of one or more schools.

    :var int ~.id: unit ID
    :var str ~.code: unit code (school code) - often 6 digits
    :var str ~.name: unit full name
    :var str ~.short_name: unit short name
    :var str ~.display_name: unit display name
    :var str ~.address: unit address (location)
    :var str ~.rest_url: unit data's API base URL
    """

    id: int = IntegerField(key="Id")
    code: str = StringField(key="Symbol")
    name: str = StringField(key="Name")
    short_name: str = StringField(key="Short")
    display_name: str = StringField(key="DisplayName")
    rest_url: str = StringField(key="RestURL")
    address: str = StringField(key="Address", required=False)



# -*- coding: utf-8 -*-

from datetime import date, datetime, time

from related import DateField, IntegerField, TimeField, immutable

from .._endpoints import DATA_INTERNAL_TIME
from ._serializable import Serializable


@immutable
class DateTime(Serializable):
    """A date-time object used for representing points in time.

    :var int ~.timestamp: number of millis since the Unix epoch
    :var `datetime.date` ~.date: a date object
    :var `datetime.time` ~.time: a time object
    """

    timestamp: int = IntegerField(key="Timestamp")
    date: date = DateField(key="Date")
    time: time = TimeField(key="Time")

    @property
    def date_time(self) -> datetime:
        """Combine the date and time of this object.

        :rtype: :class:`datetime.datetime`
        """
        return datetime.combine(self.date, self.time)

    def __str__(self) -> str:
        return self.date_time.strftime("%Y-%m-%d %H:%m:%S")

    @classmethod
    async def get(cls, api, **kwargs) -> "DateTime":
        """
        :rtype: :class:`~vulcan.model.DateTime`
        """
        return await api.helper.get_object(DateTime, DATA_INTERNAL_TIME, *kwargs)



# -*- coding: utf-8 -*-
from datetime import date

from related import DateField, IntegerField, immutable

from .._endpoints import DATA_LUCKY_NUMBER
from ..model import Serializable


@immutable
class LuckyNumber(Serializable):
    """A lucky number for the specified date.

    :var `datetime.date` ~.date: lucky number date
    :var int ~.number: the lucky number
    """

    date: date = DateField(key="Day")
    number: int = IntegerField(key="Number")

    @classmethod
    async def get(cls, api, day: date, **kwargs) -> "LuckyNumber":
        """
        :rtype: :class:`~vulcan.data.LuckyNumber`
        """
        return await api.helper.get_object(
            LuckyNumber,
            DATA_LUCKY_NUMBER,
            query={
                "constituentId": api.student.school.id,
                "day": day.strftime("%Y-%m-%d"),
            },
            **kwargs
        )



# -*- coding: utf-8 -*-
from typing import AsyncIterator, List, Union

from related import ChildField, IntegerField, SequenceField, StringField, immutable

from .._api_helper import FilterType
from .._endpoints import DATA_MESSAGE
from ..model import Attachment, DateTime, Serializable


@immutable
class Address(Serializable):
    """An address - "descriptor" used in the system containing the user's Login ID, his names and a hash of the data.

    :var str ~.login_id: login id
    :var str ~.address_name: address name
    :var str ~.address_hash: address hash
    :var str ~.initials: initials
    """

    login_id: int = IntegerField(key="LoginId")
    address_name: str = StringField(key="Address")
    address_hash: str = StringField(key="AddressHash")
    initials: str = StringField(key="Initials")


@immutable
class Message(Serializable):
    """A message.

    :var int ~.id: Message id
    :var str ~.subject: Subject of the message
    :var `~vulcan.hebe.model.DateTime` ~.sent_date: Date with time when the message was sent
    :var `~vulcan.hebe.model.DateTime` ~.read_date: Date with time when the message was read
    :var int ~.status: Message status
    :var str ~.content: Message content
    :var `~vulcan.data.Address` ~.sender: Sender of the message
    :var List[Address] ~.receivers: Receiver of the message
    :var List[Attachment] ~.attachments: attachments added to message
    """

    id: int = IntegerField(key="Id", required=False)
    subject: str = StringField(key="Subject", required=False)
    sent_date: DateTime = ChildField(DateTime, key="DateSent", required=False)
    read_date: DateTime = ChildField(DateTime, key="DateRead", required=False)
    status: int = IntegerField(key="Status", required=False)
    content: str = StringField(key="Content", required=False)
    sender: Address = ChildField(Address, key="Sender", required=False)
    receivers: List[Address] = SequenceField(
        Address, key="Receiver", repr=True, required=False
    )
    attachments: List[Attachment] = SequenceField(
        Attachment, key="Attachments", repr=True, required=False
    )

    @classmethod
    async def get(
        cls, api, last_sync, folder, **kwargs
    ) -> Union[AsyncIterator["Message"], List[int]]:
        """
        :rtype: Union[AsyncIterator[:class:`~vulcan.data.Message`], List[int]]
        """
        data = await api.helper.get_list(
            DATA_MESSAGE,
            FilterType.BY_PERSON,
            last_sync=last_sync,
            folder=folder,
            **kwargs
        )

        for message in data:
            yield Message.load(message)



# -*- coding: utf-8 -*-

from related import IntegerField, StringField, immutable

from ._serializable import Serializable


@immutable
class TeamClass(Serializable):
    """A school class.

    :var int ~.id: class ID
    :var str ~.key: class's key (UUID)
    :var str ~.display_name: class's display name
    :var str ~.symbol: class's symbol (e.g. a letter after the level, "C" in "6C")
    """

    id: int = IntegerField(key="Id")
    key: str = StringField(key="Key")
    display_name: str = StringField(key="DisplayName")
    symbol: str = StringField(key="Symbol")


@immutable
class TeamVirtual(Serializable):
    """A virtual team, i.e. a part of the school class. Often called
    a "distribution" of the class.

    :var int ~.id: team ID
    :var str ~.key: team's key (UUID)
    :var str ~.shortcut: team's short name
    :var str ~.name: team's name
    :var str ~.part_type: type of the distribution
    """

    id: int = IntegerField(key="Id")
    key: str = StringField(key="Key")
    shortcut: str = StringField(key="Shortcut")
    name: str = StringField(key="Name")
    part_type: str = StringField(key="PartType")



# -*- coding: utf-8 -*-
from typing import AsyncIterator, List, Union

from related import ChildField, FloatField, IntegerField, StringField, immutable

from .._api_helper import FilterType
from .._endpoints import DATA_GRADE
from ..model import DateTime, Period, Serializable, Subject, Teacher


@immutable
class GradeCategory(Serializable):
    """A base grade category. Represents a generic type, like an exam, a short test,
    a homework or other ("current") grades.

    :var int ~.id: grade category's ID
    :var str ~.name: grade category's name
    :var str ~.code: grade category's code (e.g. short name or abbreviation)
    """

    id: int = IntegerField(key="Id")
    name: str = StringField(key="Name")
    code: str = StringField(key="Code")


@immutable
class GradeColumn(Serializable):
    """A grade column. Represents a topic which a student
    may get a grade from (e.g. a single exam, short test, homework).

    :var int ~.id: grade column's ID
    :var str ~.key: grade column's key (UUID)
    :var int ~.period_id: ID of the period when the grade is given
    :var str ~.name: grade column's name (description)
    :var str ~.code: grade column's code (e.g. short name or abbreviation)
    :var str ~.group: unknown, yet
    :var int ~.number: unknown, yet
    :var int ~.weight: weight of this column's grades
    :var `~vulcan.model.Subject` ~.subject: the subject from which
        grades in this column are given
    :var `~vulcan.data.GradeCategory` ~.category: category (base type)
        of grades in this column
    :var `~vulcan.model.Period` ~.period: a resolved period of this grade
    """

    id: int = IntegerField(key="Id")
    key: str = StringField(key="Key")
    period_id: int = IntegerField(key="PeriodId")
    name: str = StringField(key="Name")
    code: str = StringField(key="Code")
    number: int = IntegerField(key="Number")
    weight: float = FloatField(key="Weight")
    subject: Subject = ChildField(Subject, key="Subject")
    group: str = StringField(key="Group", required=False)
    category: GradeCategory = ChildField(GradeCategory, key="Category", required=False)

    period: Period = ChildField(Period, key="Period", required=False)


@immutable
class Grade(Serializable):
    """A grade.

    :var int ~.id: grade's ID
    :var str ~.key: grade's key (UUID)
    :var int ~.pupil_id: the related pupil's ID
    :var str ~.content_raw: grade's content (with comment)
    :var str ~.content: grade's content (without comment)
    :var `~vulcan.model.DateTime` ~.date_created: grade's creation date
    :var `~vulcan.model.DateTime` ~.date_modified: grade's modification date
        (may be the same as ``date_created`` if it was never modified)
    :var `~vulcan.model.Teacher` ~.teacher_created: the teacher who added
        the grade
    :var `~vulcan.model.Teacher` ~.teacher_modified: the teacher who modified
        the grade
    :var `~vulcan.data.GradeColumn` ~.column: grade's column
    :var float ~.value: grade's value, may be `None` if 0.0
    :var str ~.comment: grade's comment, visible in parentheses in ``content_raw``
    :var float ~.numerator: for point grades: the numerator value
    :var float ~.denominator: for point grades: the denominator value
    """

    id: int = IntegerField(key="Id")
    key: str = StringField(key="Key")
    pupil_id: int = IntegerField(key="PupilId")
    content_raw: str = StringField(key="ContentRaw")
    content: str = StringField(key="Content")
    date_created: DateTime = ChildField(DateTime, key="DateCreated")
    date_modified: DateTime = ChildField(DateTime, key="DateModify")
    teacher_created: Teacher = ChildField(Teacher, key="Creator")
    teacher_modified: Teacher = ChildField(Teacher, key="Modifier")
    column: GradeColumn = ChildField(GradeColumn, key="Column")
    value: float = FloatField(key="Value", required=False)
    comment: str = StringField(key="Comment", required=False)
    numerator: float = FloatField(key="Numerator", required=False)
    denominator: float = FloatField(key="Denominator", required=False)

    @classmethod
    async def get(
        cls, api, last_sync, deleted, **kwargs
    ) -> Union[AsyncIterator["Grade"], List[int]]:
        """
        :rtype: Union[AsyncIterator[:class:`~vulcan.data.Grade`], List[int]]
        """
        data = await api.helper.get_list(
            DATA_GRADE,
            FilterType.BY_PUPIL,
            deleted=deleted,
            last_sync=last_sync,
            **kwargs
        )

        for grade in data:
            grade["Column"]["Period"] = api.student.period_by_id(
                grade["Column"]["PeriodId"]
            ).as_dict
            yield Grade.load(grade)



# -*- coding: utf-8 -*-

from aenum import Enum, unique
from related import ChildField, IntegerField, StringField, immutable

from ._serializable import Serializable


@unique
class Gender(Enum):
    """Student gender"""

    WOMAN = False
    MAN = True


@immutable
class Pupil(Serializable):
    """A class containing the student's data.

    :var int ~.id: pupil's ID
    :var int ~.login_id: pupil's account login ID
    :var str ~.login_value: pupil's account login name (email/username)
    :var str ~.first_name: student's first name
    :var str ~.second_name: student's second name, optional
    :var str ~.last_name: student's last name / surname
    :var `~vulcan.model.Gender` ~.gender: student's gender
    """

    id: int = IntegerField(key="Id")
    login_id: int = IntegerField(key="LoginId")
    first_name: str = StringField(key="FirstName")
    last_name: str = StringField(key="Surname")
    gender: Gender = ChildField(Gender, key="Sex")
    second_name: str = StringField(key="SecondName", required=False)
    login_value: str = StringField(key="LoginValue", required=False)



# -*- coding: utf-8 -*-
import datetime
from typing import AsyncIterator, List, Union

from related import BooleanField, ChildField, IntegerField, StringField, immutable

from .._api_helper import FilterType
from .._endpoints import DATA_TIMETABLE, DATA_TIMETABLE_CHANGES
from ..model import (
    DateTime,
    Serializable,
    Subject,
    Teacher,
    TeamClass,
    TeamVirtual,
    TimeSlot,
)


@immutable
class LessonRoom(Serializable):
    """Lesson room

    :var int ~.id: lesson room ID
    :var str ~.code: classroom code
    """

    id: int = IntegerField(key="Id")
    code: str = StringField(key="Code")


@immutable
class LessonChanges(Serializable):
    """Lesson changes

    :var int ~.id: lesson change ID
    :var int ~.type: lesson change type
    :var bool ~.code: team separation
    """

    id: int = IntegerField(key="Id")
    type: int = IntegerField(key="Type")
    separation: bool = BooleanField(key="Separation")


@immutable
class Lesson(Serializable):
    """A lesson.

    :var int ~.id: lesson's ID
    :var `~vulcan.model.DateTime` ~.date: lesson's date
    :var `~vulcan.model.TimeSlot` ~.time: lesson's time
    :var `~vulcan.data.LessonRoom` ~.room: classroom, in which is the lesson
    :var `~vulcan.model.Teacher` ~.teacher: teacher of the lesson
    :var `~vulcan.model.Teacher` ~.second_teacher: second teacher of the lesson
    :var `~vulcan.model.Subject` ~.subject: subject on the lesson
    :var str ~.event: an event happening during this lesson
    :var `~vulcan.data.LessonChanges` ~.changes: lesson changes
    :var `~vulcan.model.TeamClass` ~.team_class: the class that has the lesson
    :var str ~.pupil_alias: pupil alias
    :var `~vulcan.model.TeamVirtual` ~.group: group, that has the lesson
    :var bool ~.visible: lesson visibility (whether the timetable applies to the given student)
    """

    id: int = IntegerField(key="Id", required=False)
    date: DateTime = ChildField(DateTime, key="Date", required=False)
    time: TimeSlot = ChildField(TimeSlot, key="TimeSlot", required=False)
    room: LessonRoom = ChildField(LessonRoom, key="Room", required=False)
    teacher: Teacher = ChildField(Teacher, key="TeacherPrimary", required=False)
    second_teacher: Teacher = ChildField(
        Teacher, key="TeacherSecondary", required=False
    )
    subject: Subject = ChildField(Subject, key="Subject", required=False)
    event: str = StringField(key="Event", required=False)
    changes: LessonChanges = ChildField(LessonChanges, key="Change", required=False)
    team_class: TeamClass = ChildField(TeamClass, key="Clazz", required=False)
    pupil_alias: str = StringField(key="PupilAlias", required=False)
    group: TeamVirtual = ChildField(TeamVirtual, key="Distribution", required=False)
    visible: bool = BooleanField(key="Visible", required=False)

    @classmethod
    async def get(
        cls, api, last_sync, deleted, date_from, date_to, **kwargs
    ) -> Union[AsyncIterator["Lesson"], List[int]]:
        """
        :rtype: Union[AsyncIterator[:class:`~vulcan.data.Lesson`], List[int]]
        """
        if date_from is None:
            date_from = datetime.date.today()
        if date_to is None:
            date_to = date_from
        data = await api.helper.get_list(
            DATA_TIMETABLE,
            FilterType.BY_PUPIL,
            deleted=deleted,
            date_from=date_from,
            date_to=date_to,
            last_sync=last_sync,
            **kwargs,
        )

        for lesson in data:
            yield Lesson.load(lesson)


@immutable
class ChangedLesson(Serializable):
    """Changed lesson.

    :var int ~.id: changed lesson's ID
    :var int ~.unit_id: unit ID
    :var int ~.schedule_id: normal lesson's ID
    :var `~vulcan.model.DateTime` ~.lesson_date: lesson's date
    :var `~vulcan.model.DateTime` ~.change_date: change date
    :var `~vulcan.model.TimeSlot` ~.time: lesson's time
    :var str ~.note: change note
    :var str ~.reason: change reason
    :var `~vulcan.data.LessonRoom` ~.room: classroom, in which is the lesson
    :var `~vulcan.model.Teacher` ~.teacher: teacher of the lesson
    :var `~vulcan.model.Teacher` ~.second_teacher: second teacher of the lesson
    :var `~vulcan.model.Subject` ~.subject: subject on the lesson
    :var str ~.event: an event happening during this lesson
    :var `~vulcan.data.LessonChanges` ~.changes: lesson changes
    :var `~vulcan.model.TeamClass` ~.team_class: the class that has the lesson
    :var `~vulcan.model.TeamVirtual` ~.group: group, that has the lesson
    """

    id: int = IntegerField(key="Id", required=False)
    unit_id: int = IntegerField(key="UnitId", required=False)
    schedule_id: int = IntegerField(key="'ScheduleId': ", required=False)
    lesson_date: DateTime = ChildField(DateTime, key="LessonDate", required=False)
    note: str = StringField(key="Note", required=False)
    reason: str = StringField(key="Reason", required=False)
    time: TimeSlot = ChildField(TimeSlot, key="TimeSlot", required=False)
    room: LessonRoom = ChildField(LessonRoom, key="Room", required=False)
    teacher: Teacher = ChildField(Teacher, key="TeacherPrimary", required=False)
    second_teacher: Teacher = ChildField(
        Teacher, key="TeacherSecondary", required=False
    )
    subject: Subject = ChildField(Subject, key="Subject", required=False)
    event: str = StringField(key="Event", required=False)
    changes: LessonChanges = ChildField(LessonChanges, key="Change", required=False)
    change_date: DateTime = ChildField(DateTime, key="ChangeDate", required=False)
    team_class: TeamClass = ChildField(TeamClass, key="Clazz", required=False)
    group: TeamVirtual = ChildField(TeamVirtual, key="Distribution", required=False)

    @classmethod
    async def get(
        cls, api, last_sync, deleted, date_from, date_to, **kwargs
    ) -> Union[AsyncIterator["Lesson"], List[int]]:
        """
        :rtype: Union[AsyncIterator[:class:`~vulcan.data.ChangeLesson`], List[int]]
        """
        if date_from is None:
            date_from = datetime.date.today()
        if date_to is None:
            date_to = date_from
        data = await api.helper.get_list(
            DATA_TIMETABLE_CHANGES,
            FilterType.BY_PUPIL,
            deleted=deleted,
            date_from=date_from,
            date_to=date_to,
            last_sync=last_sync,
            **kwargs,
        )

        for lesson in data:
            yield ChangedLesson.load(lesson)



# -*- coding: utf-8 -*-
import datetime
from typing import AsyncIterator, List, Union

from related import BooleanField, ChildField, IntegerField, StringField, immutable

from .._api_helper import FilterType
from .._endpoints import DATA_ATTENDANCE
from ..model import (
    DateTime,
    Serializable,
    Subject,
    Teacher,
    TeamClass,
    TeamVirtual,
    TimeSlot,
)


@immutable
class PresenceType(Serializable):
    """Presence type

    :var int ~.id: attendance ID
    :var str ~.name: attendance name
    :var str ~.symbol: attendance symbol
    :var int ~.category_id: attendance category ID
    :var str ~.category_name: attendance category name
    :var int ~.position: attendance position
    :var bool ~.presence: presence on lesson
    :var bool ~.absence: absence on lesson
    :var bool ~.exemption: exemption from lesson
    :var bool ~.late: is late for lesson
    :var bool ~.justified: justified absence
    :var bool ~.deleted: whether the entry is deleted
    """

    id = IntegerField(key="Id")
    name: str = StringField(key="Name")
    symbol: str = StringField(key="Symbol")
    category_id: int = IntegerField(key="CategoryId")
    category_name: str = StringField(key="CategoryName")
    position: int = IntegerField(key="Position")
    presence: bool = BooleanField(key="Presence")
    absence: bool = BooleanField(key="Absence")
    exemption: bool = BooleanField(key="LegalAbsence")
    late: bool = BooleanField(key="Late")
    justified: bool = BooleanField(key="AbsenceJustified")
    deleted: bool = BooleanField(key="Removed")


@immutable
class Attendance(Serializable):
    """Attendance.

    :var int ~.lesson_id: lesson ID
    :var int ~.id: attendance ID
    :var int ~.lesson_number: lesson number
    :var str ~.global_key: attendance global key
    :var int ~.lesson_class_id: lesson class ID
    :var str ~.global_key: lesson class global key
    :var bool ~.calculate_presence: does it count for absences
    :var bool ~.replacement: os it replaced
    :var `~vulcan.model.Subject` ~.subject: subject of the lesson
    :var str ~.topic: topic of the lesson
    :var `~vulcan.model.Teacher` ~.teacher: teacher of the lesson
    :var `~vulcan.model.Teacher` ~.second_teacher: second teacher of the lesson
    :var `~vulcan.model.Teacher` ~.main_teacher: pupil main teacher
    :var `~vulcan.model.TeamClass` ~.team_class: the class that had lesson
    :var str ~.class_alias: class short name
    :var `~vulcan.model.DateTime` ~.date: lesson's date
    :var `~vulcan.model.TimeSlot` ~.time: lesson's time
    :var `~vulcan.model.DateTime` ~.date_modified: attendance modification date, if not modified it is created date
    :var int ~.id: aux presence ID
    :var str ~.justification_status: attendance justification status
    :var `~vulcan.data.PresenceType` ~.presence_type: presence type
    :var str ~.note: attendance note
    :var str ~.public_resources: attendance public resources
    :var str ~.remote_resources: attendance remote resources
    :var `~vulcan.model.TeamVirtual` ~.group: group, that has the lesson
    :var bool ~.visible: attendance visibility

    """

    lesson_id: int = IntegerField(key="LessonId")
    id: int = IntegerField(key="Id")
    lesson_number: int = IntegerField(key="LessonNumber")
    global_key: str = StringField(key="GlobalKey")
    lesson_class_id: int = IntegerField(key="LessonClassId")
    lesson_class_global_key: str = StringField(key="LessonClassGlobalKey")
    calculate_presence: bool = BooleanField(key="CalculatePresence")
    replacement: bool = BooleanField(key="Replacement")
    subject: Subject = ChildField(Subject, key="Subject", required=False)
    topic: str = StringField(key="Topic", required=False)
    teacher: Teacher = ChildField(Teacher, key="TeacherPrimary", required=False)
    second_teacher: Teacher = ChildField(
        Teacher, key="TeacherSecondary", required=False
    )
    main_teacher: Teacher = ChildField(Teacher, key="TeacherMod", required=False)
    team_class: TeamClass = ChildField(TeamClass, key="Clazz", required=False)
    class_alias: str = StringField(key="GroupDefinition", required=False)
    date: DateTime = ChildField(DateTime, key="Day", required=False)
    time: TimeSlot = ChildField(TimeSlot, key="TimeSlot", required=False)
    date_modified: DateTime = ChildField(DateTime, key="DateModify", required=False)
    aux_presence_id: int = IntegerField(key="AuxPresenceId", required=False)
    justification_status: str = StringField(key="JustificationStatus", required=False)
    presence_type: PresenceType = ChildField(
        PresenceType, key="PresenceType", required=False
    )
    note: str = StringField(key="Note", required=False)
    public_resources: str = StringField(key="PublicResources", required=False)
    remote_resources: str = StringField(key="RemoteResources", required=False)
    group: TeamVirtual = ChildField(TeamVirtual, key="Distribution", required=False)
    visible = BooleanField(key="Visible", required=False)

    @classmethod
    async def get(
        cls, api, last_sync, deleted, date_from, date_to, **kwargs
    ) -> Union[AsyncIterator["Attendance"], List[int]]:
        """
        :rtype: Union[AsyncIterator[:class:`~vulcan.data.Attendance`], List[int]]
        """
        if date_from is None:
            date_from = datetime.date.today()
        if date_to is None:
            date_to = date_from
        date_to = date_to + datetime.timedelta(
            days=1
        )  # Vulcan requires the date_to to be one greater the date it is supposed to be
        data = await api.helper.get_list(
            DATA_ATTENDANCE,
            FilterType.BY_PUPIL,
            deleted=deleted,
            date_from=date_from,
            date_to=date_to,
            last_sync=last_sync,
            **kwargs
        )

        for attendance in data:
            yield Attendance.load(attendance)



# -*- coding: utf-8 -*-

from ._account import Account
from ._client import Vulcan
from ._exceptions import (
    ExpiredTokenException,
    InvalidPINException,
    InvalidSignatureValuesException,
    InvalidSymbolException,
    InvalidTokenException,
    UnauthorizedCertificateException,
    VulcanAPIException,
)
from ._keystore import Keystore

__version__ = "2.1.1"
__doc__ = "Unofficial API for UONET+ e-register"

__all__ = [
    "Vulcan",
    "Keystore",
    "Account",
    "ExpiredTokenException",
    "InvalidPINException",
    "InvalidSignatureValuesException",
    "InvalidSymbolException",
    "InvalidTokenException",
    "UnauthorizedCertificateException",
    "VulcanAPIException",
]



# -*- coding: utf-8 -*-

from related import IntegerField, StringField, immutable

from ._serializable import Serializable


@immutable
class Teacher(Serializable):
    """A teacher or other school employee.

    :var int ~.id: teacher ID
    :var str ~.name: teacher's name
    :var str ~.surname: teacher's surname
    :var str ~.display_name: teacher's display name
    """

    id: int = IntegerField(key="Id")
    name: str = StringField(key="Name")
    surname: str = StringField(key="Surname")
    display_name: str = StringField(key="DisplayName")



# -*- coding: utf-8 -*-
from datetime import date, datetime
from enum import Enum, unique

from ._endpoints import DATA_BY_PERIOD, DATA_BY_PERSON, DATA_BY_PUPIL, DATA_ROOT


@unique
class FilterType(Enum):
    BY_PUPIL = 0
    BY_PERSON = 1
    BY_PERIOD = 2
    BY_LOGIN_ID = None

    def get_endpoint(self):
        if self == FilterType.BY_PUPIL:
            return DATA_BY_PUPIL
        elif self == FilterType.BY_PERSON:
            return DATA_BY_PERSON
        elif self == FilterType.BY_PERIOD:
            return DATA_BY_PERIOD
        else:
            return None


class ApiHelper:
    def __init__(self, api):
        self._api = api

    async def get_list(
        self,
        endpoint: str,
        filter_type: FilterType,
        deleted: bool = False,
        date_from: date = None,
        date_to: date = None,
        last_sync: datetime = None,
        folder: int = None,
        params: dict = None,
        **kwargs,
    ) -> list:
        if not self._api.student:
            raise AttributeError("No student is selected.")
        if deleted:
            raise NotImplementedError(
                "Getting deleted data IDs is not implemented yet."
            )
        if filter_type and filter_type != FilterType.BY_LOGIN_ID:
            url = "{}/{}/{}".format(DATA_ROOT, endpoint, filter_type.get_endpoint())
        else:
            url = "{}/{}".format(DATA_ROOT, endpoint)
        query = {}
        account = self._api.account
        student = self._api.student
        period = self._api.period

        if filter_type == FilterType.BY_PUPIL:
            query["unitId"] = student.unit.id
            query["pupilId"] = student.pupil.id
            query["periodId"] = period.id
        elif (
            filter_type == FilterType.BY_PERSON or filter_type == FilterType.BY_LOGIN_ID
        ):
            query["loginId"] = account.login_id
        elif filter_type == FilterType.BY_PERIOD:
            query["periodId"] = period.id
            query["pupilId"] = student.pupil.id

        if date_from:
            query["dateFrom"] = date_from.strftime("%Y-%m-%d")
        if date_to:
            query["dateTo"] = date_to.strftime("%Y-%m-%d")

        if folder is not None:
            query["folder"] = folder

        query["lastId"] = "-2147483648"  # don't ask, it's just Vulcan
        query["pageSize"] = 500
        query["lastSyncDate"] = (last_sync or datetime(1970, 1, 1, 0, 0, 0)).strftime(
            "%Y-%m-%d %H:%m:%S"
        )

        if params:
            query.update(params)

        return await self._api.get(url, query, **kwargs)

    async def get_object(
        self,
        cls,
        endpoint: str,
        query: dict = None,
        **kwargs,
    ) -> object:
        url = "{}/{}".format(DATA_ROOT, endpoint)

        data = await self._api.get(url, query, **kwargs)
        return cls.load(data)



# -*- coding: utf-8 -*-

import io
import re
from os import path

from setuptools import find_packages, setup

here = path.abspath(path.dirname(__file__))

with io.open(path.join(here, "README.md"), "rt", encoding="utf8") as f:
    long_description = f.read()

with io.open(path.join(here, "vulcan/__init__.py"), "rt", encoding="utf8") as f:
    version = re.search(r"__version__ = \"(.*?)\"", str(f.read())).group(1)

setup(
    name="vulcan-api",
    version=version,
    packages=find_packages(),
    author="Kacper Ziubryniewicz",
    author_email="kapi2289@gmail.com",
    description="Nieoficjalne API do dzienniczka elektronicznego UONET+",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords=["Vulcan", "UONET+", "Dzienniczek+", "API", "e-dziennik", "hebe"],
    license="MIT",
    url="https://github.com/kapi2289/vulcan-api",
    project_urls={"Documentation": "https://vulcan-api.readthedocs.io/"},
    python_requires=">=3.6,<4.0",
    install_requires=[
        "pyopenssl",
        "uonet-request-signer-hebe",
        "pytz",
        "aenum",
        "related",
        "aiohttp",
        "cchardet",
        "aiodns",
        "yarl",
    ],
    extras_require={"testing": ["pytest", "python-dotenv"]},
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: Polish",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Education",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



# -*- coding: utf-8 -*-
from typing import AsyncIterator, List, Union

from related import IntegerField, SequenceField, StringField, immutable

from .._api_helper import FilterType
from .._endpoints import DATA_ADDRESSBOOK
from ..model import Serializable


@immutable
class Role(Serializable):
    """A role of addressee.

    :var str ~.role_name: role name
    :var int ~.role_order: role order
    :var str ~.address_name: address name
    :var str ~.address_hash: address hash
    :var str ~.first_name: recipient's first name
    :var str ~.last_name: recipient's last name
    :var str ~.initials: recipient's initials
    :var str ~.unit_symbol: recipient's unit_symbol
    :var str ~.constituent_unit_symbol: recipient's constituent unit symbol
    :var str ~.class_symbol: recipient's class symbol
    """

    role_name: str = StringField(key="RoleName")
    role_order: int = IntegerField(key="RoleOrder")
    address_name: str = StringField(key="Address")
    address_hash: str = StringField(key="AddressHash")
    first_name: str = StringField(key="Name")
    last_name: str = StringField(key="Surname")
    initials: str = StringField(key="Initials")
    unit_symbol: str = StringField(key="UnitSymbol", required=False)
    constituent_unit_symbol: str = StringField(
        key="ConstituentUnitSymbol", required=False
    )
    class_symbol: str = StringField(key="ClassSymbol", required=False)


@immutable
class Addressbook(Serializable):
    """An address book.

    :var str ~.id: recipient id
    :var str ~.login_id: recipient login id
    :var str ~.first_name: recipient's first name
    :var str ~.last_name: recipient's last name
    :var str ~.initials: recipient's initials
    :var list[Role] ~.roles: recipient's role (eg. Teacher)
    """

    id: str = StringField(key="Id")
    login_id: int = IntegerField(key="LoginId")
    first_name: str = StringField(key="Name")
    last_name: str = StringField(key="Surname")
    initials: str = StringField(key="Initials")

    roles: List[Role] = SequenceField(Role, key="Roles", repr=True)

    @classmethod
    async def get(cls, api, **kwargs) -> Union[AsyncIterator["Addressbook"], List[int]]:
        """
        :rtype: Union[AsyncIterator[:class:`~vulcan.data.Addressbook`], List[int]]
        """
        data = await api.helper.get_list(
            DATA_ADDRESSBOOK, FilterType.BY_LOGIN_ID, **kwargs
        )

        for addressbook in data:
            yield Addressbook.load(addressbook)



# -*- coding: utf-8 -*-

from related import StringField, immutable
from uonet_request_signer_hebe import generate_key_pair

from ._utils import default_device_model, get_firebase_token, log
from .model import Serializable


@immutable
class Keystore(Serializable):
    """A keystore containing of:

    - a PEM-encoded X509 certificate signed using SHA-256 with RSA algorithm
    - SHA-1 fingerprint of the certificate, represented
      as lowercase hexadecimal characters
    - a PEM-encoded PKCS#8 RSA 2048 private key

    Additionally, to use with the Vulcan API the keystore contains:

    - a Firebase Cloud Messaging token - to re-use for every request
    - a device name string, also needed for API requests

    :var str ~.certificate: a PEM-encoded certificate
    :var str ~.fingerprint: the certificate's fingerprint
    :var str ~.private_key: a PEM-encoded RSA 2048 private key
    :var str ~.firebase_token: an FCM token
    :var str ~.device_model: a device model string
    """

    certificate: str = StringField(key="Certificate")
    fingerprint: str = StringField(key="Fingerprint")
    private_key: str = StringField(key="PrivateKey")
    firebase_token: str = StringField(key="FirebaseToken")
    device_model: str = StringField(key="DeviceModel")

    @staticmethod
    async def create(
        firebase_token: str = None, device_model: str = default_device_model()
    ) -> "Keystore":
        log.info("Generating key pair...")
        keystore = Keystore(
            *generate_key_pair(),
            firebase_token if firebase_token else await get_firebase_token(),
            device_model
        )
        log.info(
            "Generated for {}, sha1: {}".format(device_model, keystore.fingerprint)
        )
        return keystore



# -*- coding: utf-8 -*-

from related import IntegerField, StringField, TimeField, immutable

from .._utils import TIME_FORMAT_H_M
from ._serializable import Serializable


@immutable
class TimeSlot(Serializable):
    """Lesson time (start-end range)

    :var int ~.id: lesson time ID
    :var `datetime.time` ~.from_: lesson start time
    :var `datetime.time` ~.to: lesson end time
    :var str ~.displayed_time: lesson's displayed time
    :var int ~.position: lesson position
    """

    id: int = IntegerField(key="Id")
    from_: TimeField = TimeField(key="Start", formatter=TIME_FORMAT_H_M)
    to: TimeField = TimeField(key="End", formatter=TIME_FORMAT_H_M)
    displayed_time: str = StringField(key="Display")
    position: int = IntegerField(key="Position")
